@@ -121,6 +121,14 @@ int R_SignalHandlers = 1;  /* Exposed in R_interface.h */
 			"pg.spi.lastoid <-function() " \
 			"{.Call(\"plr_SPI_lastoid\")}"
 #endif
+#if PG_VERSION_NUM >= 110000
+#define SPI_COMMIT_CMD \
+			"pg.spi.commit <-function() " \
+			"{.Call(\"plr_SPI_commit\")}"
+#define SPI_ROLLBACK_CMD \
+			"pg.spi.rollback <-function() " \
+			"{.Call(\"plr_SPI_rollback\")}"
+#endif
 #define SPI_DBDRIVER_CMD \
 			"dbDriver <-function(db_name)\n" \
 			"{return(NA)}"
@@ -228,12 +236,26 @@ PG_FUNCTION_INFO_V1(plr_call_handler);
 Datum
 plr_call_handler(PG_FUNCTION_ARGS)
 {
+
+#if PG_VERSION_NUM >= 110000
+	bool		nonatomic;
+#endif
 	Datum			retval;
 
+#if PG_VERSION_NUM >= 110000
+	nonatomic = fcinfo->context &&
+		IsA(fcinfo->context, CallContext) &&
+		!castNode(CallContext, fcinfo->context)->atomic;
+#endif
+
 	/* save caller's context */
 	plr_caller_context = CurrentMemoryContext;
 
+#if PG_VERSION_NUM >= 110000
+	if (SPI_connect_ext(nonatomic ? SPI_OPT_NONATOMIC : 0) != SPI_OK_CONNECT)
+#else
 	if (SPI_connect() != SPI_OK_CONNECT)
+#endif
 		elog(ERROR, "SPI_connect failed");
 	plr_SPI_context = CurrentMemoryContext;
 	MemoryContextSwitchTo(plr_caller_context);
@@ -572,6 +594,10 @@ plr_load_builtins(Oid langOid)
 		SPI_CURSOR_CLOSE_CMD,
 #if CATALOG_VERSION_NO < 201811201
 		SPI_LASTOID_CMD,
+#endif
+#if PG_VERSION_NUM >= 110000
+		SPI_COMMIT_CMD,
+		SPI_ROLLBACK_CMD,
 #endif
 		SPI_DBDRIVER_CMD,
 		SPI_DBCONN_CMD,
@@ -1319,22 +1345,7 @@ do_compile(FunctionCallInfo fcinfo,
 			}
 			typeStruct = (Form_pg_type) GETSTRUCT(typeTup);
 
-			/* Disallow pseudotype argument
-			 * note we already replaced ANYARRAY/ANYELEMENT
-			 */
-			if (typeStruct->typtype == 'p')
-			{
-				Oid		arg_typid = function->arg_typid[i];
-
-				xpfree(function->proname);
-				xpfree(function);
-				ereport(ERROR,
-						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-						 errmsg("plr functions cannot take type %s",
-								format_type_be(arg_typid))));
-			}
-
-			if (typeStruct->typrelid != InvalidOid)
+			if (typeStruct->typrelid != InvalidOid || typeStruct->typtype == 'p')
 				function->arg_is_rel[i] = 1;
 			else
 				function->arg_is_rel[i] = 0;
@@ -1647,7 +1658,7 @@ plr_convertargs(plr_function *function, Datum *arg, bool *argnull, FunctionCallI
 			else if (function->arg_is_rel[i])
 			{
 				/* for tuple args, convert to a one row data.frame */
-				CONVERT_TUPLE_TO_DATAFRAME;
+				CONVERT_TUPLE_TO_DATAFRAME(GET_ARG_VALUE(i));
 			}
 			else if (function->arg_elem[i] == InvalidOid)
 			{
@@ -1690,11 +1701,9 @@ plr_convertargs(plr_function *function, Datum *arg, bool *argnull, FunctionCallI
 			}
 			else if (function->arg_is_rel[i])
 			{
-				/* keep compiler quiet */
-				el = R_NilValue;
-
-				elog(ERROR, "Tuple arguments not supported in PL/R Window Functions");
-			}
+				/* for tuple args, convert to a one row data.frame */
+				CONVERT_TUPLE_TO_DATAFRAME(dvalue);
+ 			}
 			else if (function->arg_elem[i] == InvalidOid)
 			{
 				/* for scalar args, convert to a one row vector */
@@ -1731,16 +1740,16 @@ plr_convertargs(plr_function *function, Datum *arg, bool *argnull, FunctionCallI
 
 		if (plr_is_unbound_frame(winobj))
 		{
-
 			SEXP lst;
 			if (0 == current_row)
 			{
 				lst = PROTECT(allocVector(VECSXP, function->nargs));
 				for (i = 0; i < function->nargs; i++, t = CDR(t))
 				{
-					el = get_fn_expr_arg_stable(fcinfo->flinfo, i) ?
-						R_NilValue : pg_window_frame_get_r(winobj, i, function);
+					PROTECT(el = /* get_fn_expr_arg_stable(fcinfo->flinfo, i) ?
+						R_NilValue : */ pg_window_frame_get_r(winobj, i, function));
 					SET_VECTOR_ELT(lst, i, el);
+					UNPROTECT(1);
 					SETCAR(t, el);
 				}
 				defineVar(install(PLR_WINDOW_FRAME_NAME), lst, rho);
@@ -1761,16 +1770,16 @@ plr_convertargs(plr_function *function, Datum *arg, bool *argnull, FunctionCallI
 		else
 			for (i = 0; i < function->nargs; i++, t = CDR(t))
 			{
-				if (get_fn_expr_arg_stable(fcinfo->flinfo, i))
-					el = R_NilValue;
-				else
-				{
-					el = pg_window_frame_get_r(winobj, i, function);
-					numels = LENGTH(el);
-				}
+				PROTECT(el = /* get_fn_expr_arg_stable(fcinfo->flinfo, i) ?
+					R_NilValue : */ pg_window_frame_get_r(winobj, i, function));
+
 				SETCAR(t, el);
+				UNPROTECT(1);
 			}
 
+		/* below only works if last el <> R_NilValue (see commented out above) */
+		numels = function->nargs > 0 ? GET_LENGTH(el) : 0;
+
 		/* fnumrows */
 		SETCAR(t, ScalarInteger(numels));
 		t = CDR(t);
